@@ -1,25 +1,35 @@
-"""A helper class to run rate-limited API requests concurrently using threads.
-"""
+"""A helper class to run rate-limited API requests concurrently using threads."""
+
 import os
 import copy
 import threading
 from typing import Any
 from datetime import datetime
-from concurrent.futures import ThreadPoolExecutor, as_completed
+from concurrent.futures import ThreadPoolExecutor, as_completed, TimeoutError as FuturesTimeoutError
 import logging
 
 import openai
 import litellm
 from pyrate_limiter import Limiter, Rate, Duration
 
-from .api_tracker import APIUsageTracker  # Integrated API usage tracker
+from .api_tracker import APIUsageTracker
+from ._params import ALL_COMPLETION_PARAMS
+
 
 # OpenAI API Tier 4 has a rate limit of 10K RPM and 2M-10M TPM
 OPENAI_API_REQUESTS_PER_MINUTE = 10_000
-# OPENAI_API_TOKENS_PER_MINUTE = 10_000_000
 OPENAI_API_TOKENS_PER_MINUTE = 2_000_000
 
-DEFAULT_MAX_CONTEXT_TOKENS = 20_000  # 20K tokens context window (default)
+# Default max context window tokens fallback
+MAX_CONTEXT_TOKENS_FALLBACK_ENV_VAR = "MAX_CONTEXT_TOKENS_FALLBACK"
+try:
+    DEFAULT_MAX_CONTEXT_TOKENS = int(os.getenv(MAX_CONTEXT_TOKENS_FALLBACK_ENV_VAR, "100000"))
+except ValueError:
+    logging.getLogger(__name__).warning(
+        f"Environment variable {MAX_CONTEXT_TOKENS_FALLBACK_ENV_VAR} must be an integer. "
+        "Falling back to 100,000 tokens.")
+    DEFAULT_MAX_CONTEXT_TOKENS = 100_000
+MAX_INPUT_TOKENS_OVERRIDE_ENV_VAR = "MAX_INPUT_TOKENS_OVERRIDE"
 
 
 class APIClient:
@@ -48,16 +58,21 @@ class APIClient:
         max_requests_per_minute: int = OPENAI_API_REQUESTS_PER_MINUTE,
         max_tokens_per_minute: int = OPENAI_API_TOKENS_PER_MINUTE,
         max_workers: int = None,
+        max_delay_seconds: int = 5 * 60,
     ):
-        """
+        """Initialize the API client.
+
         Parameters
         ----------
         max_requests_per_minute : int, optional
             Maximum API requests allowed per minute. Default is OPENAI_API_RPM.
         max_tokens_per_minute : int, optional
-            Maximum tokens allowed per minute. Default is None (no token limit).
+            Maximum tokens allowed per minute.
         max_workers : int, optional
             Maximum number of worker threads. Default is min(CPU count * 20, max_rpm).
+        max_delay_seconds : int, optional
+            Maximum time in seconds that the internal rate limiter will wait to acquire
+            resources before timing out (applies to both RPM and TPM limiters). Default is 5 minutes.
         """
         self.max_requests_per_minute = max_requests_per_minute
         self.max_tokens_per_minute = max_tokens_per_minute
@@ -72,32 +87,30 @@ class APIClient:
 
         # Set up rate limiter using pyrate-limiter
         limiter_config = {
-            "max_delay": 120_000,    # 2 minutes
+            "max_delay": max_delay_seconds * 1000,  # milliseconds
         }
 
-        # Set up RPM limiter
+        # RPM limiter
         self._rpm_limiter = None
         if self.max_requests_per_minute is not None:
             requests_limit = Rate(self.max_requests_per_minute, Duration.MINUTE)
             self._rpm_limiter = Limiter(requests_limit, **limiter_config)
 
-        # Set up TPM limiter
+        # TPM limiter
         self._tpm_limiter = None
         if self.max_tokens_per_minute is not None:
             tokens_limit = Rate(self.max_tokens_per_minute, Duration.MINUTE)
             self._tpm_limiter = Limiter(tokens_limit, **limiter_config)
 
-        # Set up logger
+        # Logger
         self._logger = logging.getLogger(__name__)
-        self._logged_msgs = set()   # To avoid logging duplicate messages over and over
+        self._logged_msgs = set()
 
         # Remove handlers from litellm logger so messages propagate to root logger
-        # > This way all LiteLLM logs will follow the settings of the root logger
-        # > (e.g. log level, format, etc.)
         litellm_logger = logging.getLogger("LiteLLM")
         litellm_logger.handlers.clear()
 
-        # Create an APIUsageTracker instance and set it up with LiteLLM.
+        # Usage tracker
         self._tracker = APIUsageTracker()
         self._tracker.set_up_litellm_cost_tracking()
 
@@ -117,14 +130,24 @@ class APIClient:
     @property
     def tracker(self) -> APIUsageTracker:
         """
-        APIUsageTracker: The API usage tracker instance.
+        The API usage tracker instance.
+
+        Returns
+        -------
+        llm_api_client.api_tracker.APIUsageTracker
+            The API usage tracker.
         """
         return self._tracker
 
     @property
     def history(self) -> list[dict]:
         """
-        list[dict]: The history of requests and responses.
+        The history of requests and responses.
+
+        Returns
+        -------
+        list[dict]
+            A list of request/response entries.
         """
         return self._history
 
@@ -160,57 +183,45 @@ class APIClient:
             A list of response objects returned by the API function calls.
             If a request fails, the corresponding response will be None.
         """
-        # Create a list to store results, maintaining the order of requests
+        # Short-circuit: no requests
+        if not requests:
+            return []
+
         responses = [None] * len(requests)
 
         # Override max_workers if provided
         max_workers = min(max_workers or self._max_workers, len(requests))
 
         with ThreadPoolExecutor(max_workers=max_workers) as executor:
-            # Submit all requests to the executor
             future_to_request_idx = {
                 executor.submit(self._rate_limited_request, request=request, sanitize=sanitize): idx
                 for idx, request in enumerate(requests)
             }
 
             try:
-                # Collect results as they complete
                 for future in as_completed(future_to_request_idx, timeout=timeout):
                     request_idx = future_to_request_idx[future]
-
                     try:
                         request_response = future.result()
-                        # Store responses in the same order as the requests
                         responses[request_idx] = request_response
-
-                    # Handle generic API errors
                     except openai.APIError as e:
                         status_code = getattr(e, 'status_code', None)
                         message = getattr(e, 'message', None)
                         llm_provider = getattr(e, 'llm_provider', None)
                         self._logger.error(
-                            "Request generated an APIError: "
-                            "status_code=%s; message=%s; llm_provider=%s",
+                            "Request generated an APIError: status_code=%s; message=%s; llm_provider=%s",
                             status_code, message, llm_provider)
-                        responses[request_idx] = None   # Explicitly set failed responses to None
-
-                    # Catch all other exceptions thrown by child threads
+                        responses[request_idx] = None
                     except Exception as e:
                         self._logger.error("Request generated an exception: %s", e)
-                        responses[request_idx] = None   # Explicitly set failed responses to None
-
-            # Handle timeout from as_completed
-            except TimeoutError as e:
+                        responses[request_idx] = None
+            except FuturesTimeoutError as e:
                 incomplete_indices = [idx for idx, resp in enumerate(responses) if resp is None]
                 self._logger.error(
-                    "Timeout reached after %s seconds. %d/%d requests did not complete; "
-                    "error=%s",
+                    "Timeout reached after %s seconds. %d/%d requests did not complete; error=%s",
                     timeout, len(incomplete_indices), len(requests), e)
-                # Note: responses array already initialized with None values for incomplete requests
 
-        # Save all responses to API client history
         self.__save_history(requests=requests, responses=responses)
-
         return responses
 
     def make_requests_with_retries(
@@ -246,14 +257,11 @@ class APIClient:
         list[object]
             A list of response objects returned by the API function calls.
         """
-        # **Break case**: Check if we've exceeded max_retries
         if current_retry > max_retries:
             self._logger.error(
-                f"Exceeded max_retries ({max_retries}) for {len(requests)} requests; "
-                "returning None responses for all requests")
+                f"Exceeded max_retries ({max_retries}) for {len(requests)} requests; returning None responses")
             return [None] * len(requests)
 
-        # **Base case**: Make the API calls
         responses = self.make_requests(
             requests=requests,
             max_workers=max_workers,
@@ -261,7 +269,6 @@ class APIClient:
             timeout=timeout,
         )
 
-        # **Recursive case**: Gather failed requests and retry them
         failed_requests = []
         failed_requests_og_indices = []
         for idx, response in enumerate(responses):
@@ -269,13 +276,10 @@ class APIClient:
                 self._logger.warning(
                     f"Request with idx={idx} failed; will be retried; "
                     f"Current retry: {current_retry + 1}/{max_retries};")
-
                 failed_requests.append(requests[idx])
                 failed_requests_og_indices.append(idx)
 
-        # If there are failed requests, retry them
         if failed_requests:
-            # Recursively retry failed requests
             failed_requests_responses = self.make_requests_with_retries(
                 requests=failed_requests,
                 max_workers=max_workers,
@@ -285,7 +289,6 @@ class APIClient:
                 current_retry=current_retry + 1,
             )
 
-            # Update the responses in the order of the original requests
             for idx_in_failed_requests, idx_in_original_requests in enumerate(failed_requests_og_indices):
                 responses[idx_in_original_requests] = failed_requests_responses[idx_in_failed_requests]
 
@@ -302,15 +305,11 @@ class APIClient:
         sanitized_request : dict
             A dictionary containing parsed and filtered request parameters.
         """
-        # 1. Remove unsupported parameters
         sanitized_request = self.remove_unsupported_params(request)
-
-        # 2. Truncate the request to the maximum context tokens for the model
         sanitized_request["messages"] = self.truncate_to_max_context_tokens(
             messages=sanitized_request["messages"],
             model=sanitized_request["model"],
         )
-
         return sanitized_request
 
     def remove_unsupported_params(self, request: dict) -> dict:
@@ -324,42 +323,41 @@ class APIClient:
             A dictionary containing the provided request with all unsupported
             parameters removed.
         """
-        # Make a copy of the request to avoid mutating the original
         request = copy.deepcopy(request)
-
-        # Extract model and messages from request
         model = request.pop("model")
         messages = request.pop("messages")
 
-        # Sanitize general provider params
         supported_params = litellm.get_supported_openai_params(
             model=model,
             request_type="chat_completion",
         )
 
-        # Hardcoding some model-specific params
         model_specific_unsupported_params = set()
         if model.lower().startswith("openai/o"):
             model_specific_unsupported_params = {"temperature"}
 
-        # Eliminate model-specific unsupported params
-        supported_params = [
-            p for p in supported_params
-            if p not in model_specific_unsupported_params
-        ]
+        supported_params = [p for p in supported_params if p not in model_specific_unsupported_params]
 
-        # Construct dict of supported kwargs
         supported_kwargs = {k: v for k, v in request.items() if k in supported_params}
+        provider_specific_kwargs = {k: v for k, v in request.items() if k not in ALL_COMPLETION_PARAMS}
 
-        # Log unsupported kwargs
-        unsupported_kwargs = {k: v for k, v in request.items() if k not in supported_params}
+        if provider_specific_kwargs:
+            msg = f"Provider-specific parameters for model='{model}' in API request: {provider_specific_kwargs}."
+            if msg not in self._logged_msgs:
+                self._logger.info(msg)
+                self._logged_msgs.add(msg)
+
+        unsupported_kwargs = {
+            k: v for k, v in request.items()
+            if (k not in supported_params and k not in provider_specific_kwargs)
+        }
         if unsupported_kwargs:
             msg = f"Unsupported parameters for model='{model}' in API request: {unsupported_kwargs}."
             if msg not in self._logged_msgs:
                 self._logger.error(msg)
                 self._logged_msgs.add(msg)
 
-        return {"model": model, "messages": messages, **supported_kwargs}
+        return {"model": model, "messages": messages, **supported_kwargs, **provider_specific_kwargs}
 
     def truncate_to_max_context_tokens(
         self,
@@ -382,59 +380,77 @@ class APIClient:
             less than or equal to the maximum context tokens for the model.
         """
         messages = copy.deepcopy(messages)
-        request_char_len = sum(len(msg["content"]) for msg in messages)
-        request_tok_len = self.count_messages_tokens(messages, model=model)
         max_tokens = self.get_max_context_tokens(model)
 
-        # Easy heuristic: iteratively drop the last 10% of the prompt
-        while request_tok_len > max_tokens:
-            # Estimate the ratio of chars to keep by comparing request_tok_len and max_tokens
-            ratio_of_chars_to_keep = min(
-                0.9,  # always drop at least 10%
-                (max_tokens / request_tok_len) - 0.05,
-            )
+        def total_chars(msgs: list[dict]) -> int:
+            return sum(len(m["content"]) for m in msgs)
 
-            # Log a warning if we're dropping more than 50% of the message
-            if ratio_of_chars_to_keep < 0.5:
+        request_tok_len = self.count_messages_tokens(messages, model=model)
+        request_char_len = total_chars(messages)
+
+        while request_tok_len > max_tokens and messages:
+            chars_per_token = max(1.0, request_char_len / max(request_tok_len, 1))
+            tokens_to_drop = max(1, request_tok_len - max_tokens)
+            approx_chars_to_drop = int(tokens_to_drop * chars_per_token)
+
+            max_step = max(1, int(request_char_len * 0.15))
+            num_chars_to_drop = min(approx_chars_to_drop, max_step)
+
+            drop_fraction = num_chars_to_drop / max(request_char_len, 1)
+            if drop_fraction > 0.5:
                 self._logger.warning(
-                    f"Dropping {1 - ratio_of_chars_to_keep} of the message due to "
-                    f"token limit; request_char_len={request_char_len}, "
-                    f"max_tokens={max_tokens}, request_tok_len={request_tok_len};")
+                    f"Dropping {drop_fraction} of the message due to token limit; "
+                    f"request_char_len={request_char_len}, max_tokens={max_tokens}, request_tok_len={request_tok_len};")
 
-            # Ensure we always keep at least 5% of the message
-            ratio_of_chars_to_keep = max(0.05, ratio_of_chars_to_keep)
-            assert 0 < ratio_of_chars_to_keep < 1
+            self._drop_chars_from_messages(messages, num_chars_to_drop)
 
-            # Compute number of chars to drop
-            num_chars_to_drop = int(request_char_len * (1 - ratio_of_chars_to_keep))
-            assert num_chars_to_drop > 0
-
-            # Drop `num_chars_to_drop` chars from the last message
-            if len(messages[-1]["content"]) > num_chars_to_drop:
-                messages[-1]["content"] = messages[-1]["content"][: -num_chars_to_drop]
-            else:
-                # If the last message is shorter than `num_chars_to_drop`, drop the entire message
-                messages.pop()
-
-                self._logger.warning(
-                    f"Could not truncate {num_chars_to_drop} chars from the last message; "
-                    f"dropping entire message instead...")
-
-            # Re-count tokens
             request_tok_len = self.count_messages_tokens(messages, model=model)
-            request_char_len = sum(len(msg["content"]) for msg in messages)
+            request_char_len = total_chars(messages)
 
         return messages
 
+    def _select_trim_index(self, messages: list[dict]) -> int:
+        """Select the index of the message to trim, preferring non-system messages."""
+        idx = len(messages) - 1
+        while idx >= 0 and messages[idx].get("role") == "system":
+            idx -= 1
+        if idx < 0:
+            idx = len(messages) - 1
+        return idx
+
+    def _drop_chars_from_messages(self, messages: list[dict], num_chars_to_drop: int) -> None:
+        """Drop a number of characters from the tail of messages, skipping system messages when possible."""
+        remaining = num_chars_to_drop
+        while remaining > 0 and messages:
+            idx = self._select_trim_index(messages)
+            current_len = len(messages[idx]["content"])
+            if current_len > remaining:
+                messages[idx]["content"] = messages[idx]["content"][:-remaining]
+                remaining = 0
+            else:
+                remaining -= current_len
+                messages.pop(idx)
+
     def get_max_context_tokens(self, model: str) -> int:
         """Get the maximum context tokens for a model."""
+        # Allow env override to take precedence over LiteLLM-reported value
+        override_value_str = os.getenv(MAX_INPUT_TOKENS_OVERRIDE_ENV_VAR)
+        if override_value_str:
+            try:
+                return int(override_value_str)
+            except ValueError:
+                self._logger.warning(
+                    f"Environment variable {MAX_INPUT_TOKENS_OVERRIDE_ENV_VAR} must be an integer. Ignoring override.")
         try:
             model_info = litellm.get_model_info(model)
-            return model_info.get("max_input_tokens", DEFAULT_MAX_CONTEXT_TOKENS)
+            max_tokens = model_info.get("max_input_tokens")
+            if max_tokens is None:
+                raise ValueError("max_input_tokens not provided by litellm")
+            return max_tokens
         except Exception as e:
             self._logger.warning(
-                f"Could not get model info from litellm: {e}. "
-                f"Using a conservative default of {DEFAULT_MAX_CONTEXT_TOKENS} tokens.")
+                f"Could not get max context tokens from litellm: {e}. "
+                f"Using fallback default of {DEFAULT_MAX_CONTEXT_TOKENS} tokens.")
             return DEFAULT_MAX_CONTEXT_TOKENS
 
     def count_messages_tokens(
@@ -461,8 +477,7 @@ class APIClient:
             The number of tokens in the messages.
         """
         msgs_char_len = sum(
-            len(msg["content"])
-            for msg in messages
+            len(msg["content"]) for msg in messages
             if (msg is not None and "content" in msg and msg["content"] is not None)
         )
 
@@ -470,20 +485,21 @@ class APIClient:
             f"Counting tokens for model={model} and messages of char length={msgs_char_len}")
 
         try:
-            # Use ThreadPoolExecutor to run token counting with a timeout
             with ThreadPoolExecutor(max_workers=1) as executor:
                 future = executor.submit(litellm.token_counter, model=model, messages=messages)
                 return future.result(timeout=timeout)
-
+        except FuturesTimeoutError:
+            self._logger.error(
+                f"Token counting timed out after {timeout} seconds. Using a rough approximation.")
+            return msgs_char_len // 3
         except TimeoutError:
             self._logger.error(
                 f"Token counting timed out after {timeout} seconds. Using a rough approximation.")
-            return msgs_char_len // 3  # Rough approximation
-
+            return msgs_char_len // 3
         except Exception as e:
             self._logger.warning(
                 f"Could not count tokens using litellm: {e}. Using a rough approximation.")
-            return msgs_char_len // 3  # Rough approximation
+            return msgs_char_len // 3
 
     def _rate_limited_request(
         self,
@@ -509,25 +525,18 @@ class APIClient:
         response : object
             The response object returned by the API function call.
         """
-        # Sanitize request if requested
         if sanitize:
             request = self.sanitize_completion_request(request)
 
-        # Block until the rate limit is available
         self._acquire_rate_limited_resources(request=request)
 
-        # Log the API request parameters
         self._logger.debug(
-            f"Thread {threading.get_ident()} making API call with "
-            f"parameters: {request}")
+            f"Thread {threading.get_ident()} making API call with parameters: {request}")
 
         try:
-            # Make the API call
             response = self.api_call(**request)
             self._logger.debug(
-                f"Thread {threading.get_ident()} completed API call with "
-                f"response: {response}")
-
+                f"Thread {threading.get_ident()} completed API call with response: {response}")
         except Exception as e:
             self._logger.error(
                 f"Thread {threading.get_ident()} failed API call with error: {e}")
@@ -535,7 +544,7 @@ class APIClient:
 
         return response
 
-    def _acquire_rate_limited_resources(self, *, request: dict) -> None:
+    def _acquire_rate_limited_resources(self, *, request: dict) -> None:    # noqa: C901
         """Wait for the rate limit to be available and acquire necessary resources.
 
         This function blocks the current thread until rate limits allow the request
@@ -546,34 +555,40 @@ class APIClient:
         request : dict
             The API request parameters containing model and messages.
         """
-        # Count tokens in the request to know how many token resources to consume
         model = request.get("model")
         messages = request.get("messages", [])
-
-        # Get token count for this request
         token_count = self.count_messages_tokens(messages, model=model)
 
-        # Log that we're waiting for rate limit
         thread_id = threading.get_ident()
         self._logger.debug(
-            f"Thread {thread_id} waiting for rate limit: "
-            f"request={1}, tokens={token_count}"
-        )
+            f"Thread {thread_id} waiting for rate limit: request={1}, tokens={token_count}")
 
-        # Block until we can acquire the rate limit resources
-        # NOTE: Acquiring each resource separately can lead to a RACE CONDITION
-        # where we acquire the RPM but not the TPM, or vice versa.
-        # However, it won't lead to a DEADLOCK since resources get released over
-        # time -- but it will lead to suboptimal resource utilization.
+        if self._rpm_limiter is None and self._tpm_limiter is None:
+            return
+
+        # Acquire RPM first (single unit)
+        rpm_lock_acquired = True
         if self._rpm_limiter is not None:
-            self._rpm_limiter.try_acquire("api_calls", weight=1)
-        if self._tpm_limiter is not None:
-            self._tpm_limiter.try_acquire("tokens", weight=token_count)
+            rpm_lock_acquired = False
+            rpm_lock_acquired = self._rpm_limiter.try_acquire("api_calls", 1)
 
-        self._logger.debug(
-            f"Thread {thread_id} acquired rate limit resources: "
-            f"request={1}, tokens={token_count}"
-        )
+        # Acquire TPM tokens in one request; rely on pyrate-limiter to handle weights
+        tpm_lock_acquired = True
+        if rpm_lock_acquired and self._tpm_limiter is not None and token_count > 0:
+            tpm_lock_acquired = False
+            tpm_lock_acquired = self._tpm_limiter.try_acquire("tokens", weight=token_count)
+
+        if rpm_lock_acquired and tpm_lock_acquired:
+            self._logger.debug(
+                f"Thread {thread_id} acquired rate limit resources: request={1}, tokens={token_count}")
+        else:
+            # Do not proceed with the API call if we could not acquire within max_delay
+            message = (
+                f"Thread {thread_id} FAILED to acquire rate limit resources: request={1}, tokens={token_count}; "
+                f"rpm_lock_acquired={rpm_lock_acquired}, tpm_lock_acquired={tpm_lock_acquired}"
+            )
+            self._logger.error(message)
+            raise RuntimeError("Rate limit acquisition failed due to max_delay constraint")
 
     def __save_history(self, *, requests: list[dict], responses: list[object]) -> None:
         """Save API requests and responses to the client's history.
@@ -592,11 +607,16 @@ class APIClient:
             """Convert a Unix timestamp to a formatted datetime string."""
             return datetime.fromtimestamp(response["created"]).strftime('%Y-%m-%d %H:%M:%S')
 
+        def get_response_content(response):
+            if response.choices and len(response.choices) > 0:
+                return response.choices[0].message.content
+            return None
+
         self._history.extend([
             {
                 "request": request,
                 "response": get_response_dict(response) if response else None,
-                "content": response.choices[0].message.content if response else None,
+                "content": get_response_content(response) if response else None,
                 "created_at": get_response_datetime(response) if response else None,
             }
             for request, response in zip(requests, responses)
