@@ -99,20 +99,22 @@ class TestAPIClient:
 
         # Request with both supported and unsupported parameters
         request = {
-            "model": "gpt-3.5-turbo",
+            "model": "openai/o3-mini",
             "messages": [{"role": "user", "content": "Hello"}],
-            "temperature": 0.7,
             "max_tokens": 100,
-            "unsupported_param": "value"  # This should be removed
+            "provider_specific_param": "value",  # This should be preserved as provider-specific
+            "temperature": 0.7,                  # This should be removed as it's not supported by o3-mini
         }
 
         # Make request with sanitization
         _ = client.make_requests([request])
 
-        # Check that unsupported_param was removed when making the request
+        # Check that:
+        # - unsupported `temperature` param was removed when making the request
+        # - provider-specific `provider_specific_param` was preserved
         sanitized_request = mock_completion.call_args[1]
-        assert "unsupported_param" not in sanitized_request
-        assert "temperature" in sanitized_request
+        assert "temperature" not in sanitized_request           # Not compatible with o3-mini
+        assert "provider_specific_param" in sanitized_request   # Unexpected kwargs are preserved as provider-specific
         assert "max_tokens" in sanitized_request
         assert "messages" in sanitized_request
 
@@ -219,6 +221,50 @@ class TestAPIClient:
         assert len(results) == n_requests
         assert mock_completion.call_count == n_requests
 
+    @patch('litellm.completion')
+    @patch('litellm.token_counter')
+    def test_max_delay_seconds_stops_blocking_early(self, mock_token_counter, mock_completion):
+        """Test that max_delay_seconds bounds how long rate limiting blocks.
+
+        With RPM=1 and two concurrent requests, the second request would normally
+        need to wait ~60s for the next slot. By setting max_delay_seconds=1, the
+        limiter should stop blocking early and allow the call to proceed without
+        waiting the full minute.
+        """
+        # Setup mocks
+        mock_token_counter.return_value = 1
+        mock_completion.return_value = MockResponse()
+
+        # Initialize client with strong RPM limit and very small max_delay
+        client = APIClient(
+            max_requests_per_minute=1,
+            max_tokens_per_minute=None,
+            max_workers=2,
+            max_delay_seconds=1,
+        )
+
+        # Prepare 2 requests that will be executed concurrently
+        requests = [
+            {"model": "gpt-3.5-turbo", "messages": [{"role": "user", "content": f"Test {i}"}]}
+            for i in range(2)
+        ]
+
+        start_time = time.time()
+        results = client.make_requests(requests, max_workers=2, sanitize=False)
+        elapsed = time.time() - start_time
+
+        # The second request should NOT block for ~60s; it should return quickly
+        # and should NOT invoke the completion due to max_delay being exceeded.
+        assert len(results) == 2
+        # Exactly one request should succeed and one should fail due to max_delay
+        num_none = sum(1 for r in results if r is None)
+        num_ok = sum(1 for r in results if r is not None)
+        assert num_none == 1 and num_ok == 1
+        # Only one API call should have been made
+        assert mock_completion.call_count == 1
+        # Ensure we didn't wait a full minute
+        assert elapsed <= 3, f"Expected requests to complete quickly due to max_delay_seconds; took {elapsed:.2f}s"
+
     @patch('litellm.completion')
     def test_request_failure_handling(self, mock_completion):
         """Test handling of failed requests."""
@@ -494,7 +540,7 @@ class TestAPIClient:
         final_token_count = client.count_messages_tokens(truncated_messages, model="gpt-3.5-turbo")
 
         # Verify the token count is now within limits
-        reasonable_lower_bound = int(max_tokens * 0.8)
+        reasonable_lower_bound = int(max_tokens * 0.9)
         assert reasonable_lower_bound <= final_token_count <= max_tokens, (
             f"Truncated message from {initial_token_count} tokens to "
             f"{final_token_count} tokens; should be between "
@@ -548,7 +594,7 @@ class TestAPIClient:
         final_token_count = client.count_messages_tokens(truncated_messages, model="gpt-3.5-turbo")
 
         # Verify the token count is now within limits
-        reasonable_lower_bound = int(max_tokens * 0.8)
+        reasonable_lower_bound = int(max_tokens * 0.9)
         assert reasonable_lower_bound <= final_token_count <= max_tokens, (
             f"Truncated message from {initial_token_count} tokens to "
             f"{final_token_count} tokens; should be between "
